#
# Copyright 2019 aiohomekit team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import socket

import pytest
from zeroconf.asyncio import AsyncServiceInfo

from aiohomekit.model.categories import Categories
from aiohomekit.model.feature_flags import FeatureFlags
from aiohomekit.zeroconf import HomeKitService


def test_simple():
    desc = {
        b"c#": b"1",
        b"id": b"00:00:01:00:00:02",
        b"md": b"unittest",
        b"s#": b"11",
        b"ci": b"5",
        b"sf": b"0",
        b"ff": b"1",
    }
    info = AsyncServiceInfo(
        "_hap._tcp.local.",
        "foo2._hap._tcp.local.",
        addresses=[socket.inet_aton("127.0.0.1")],
        port=1234,
        properties=desc,
        weight=0,
        priority=0,
    )
    svc = HomeKitService.from_service_info(info)

    assert svc.name == "foo2"
    assert svc.type == "_hap._tcp.local."
    assert svc.id == "00:00:01:00:00:02"
    assert svc.model == "unittest"
    assert svc.config_num == 1
    assert svc.state_num == 11
    assert svc.status_flags == 0
    assert svc.feature_flags == FeatureFlags.SUPPORTS_APPLE_AUTHENTICATION_COPROCESSOR
    assert svc.category == Categories.LIGHTBULB
    assert svc.address == "127.0.0.1"
    assert svc.addresses == ["127.0.0.1"]
    assert svc.port == 1234


def test_udp():
    desc = {
        b"c#": b"1",
        b"id": b"00:00:01:00:00:02",
        b"md": b"unittest",
        b"s#": b"11",
        b"ci": b"5",
        b"sf": b"0",
        b"ff": b"1",
    }
    info = AsyncServiceInfo(
        "_hap._udp.local.",
        "foo2._hap._udp.local.",
        addresses=[socket.inet_aton("127.0.0.1")],
        port=1234,
        properties=desc,
        weight=0,
        priority=0,
    )
    svc = HomeKitService.from_service_info(info)

    assert svc.name == "foo2"
    assert svc.type == "_hap._udp.local."


def test_upper_case_keys():
    desc = {
        b"C#": b"1",
        b"ID": b"00:00:01:00:00:02",
        b"MD": b"unittest",
        b"S#": b"11",
        b"CI": b"5",
        b"SF": b"0",
        b"FF": b"1",
    }
    info = AsyncServiceInfo(
        "_hap._tcp.local.",
        "foo2._hap._tcp.local.",
        addresses=[socket.inet_aton("127.0.0.1")],
        port=1234,
        properties=desc,
        weight=0,
        priority=0,
    )
    svc = HomeKitService.from_service_info(info)

    assert svc.name == "foo2"
    assert svc.type == "_hap._tcp.local."
    assert svc.id == "00:00:01:00:00:02"
    assert svc.model == "unittest"
    assert svc.config_num == 1
    assert svc.state_num == 11
    assert svc.status_flags == 0
    assert svc.feature_flags == FeatureFlags.SUPPORTS_APPLE_AUTHENTICATION_COPROCESSOR
    assert svc.category == Categories.LIGHTBULB
    assert svc.address == "127.0.0.1"
    assert svc.addresses == ["127.0.0.1"]
    assert svc.port == 1234


def test_missing_cn():
    desc = {
        b"id": b"00:00:01:00:00:02",
        b"md": b"unittest",
        b"s#": b"11",
        b"ci": b"5",
        b"sf": b"0",
        b"ff": b"1",
    }
    info = AsyncServiceInfo(
        "_hap._tcp.local.",
        "foo2._hap._tcp.local.",
        addresses=[socket.inet_aton("127.0.0.1")],
        port=1234,
        properties=desc,
        weight=0,
        priority=0,
    )
    svc = HomeKitService.from_service_info(info)

    assert svc.config_num == 0


def test_missing_sn():
    desc = {
        b"c#": b"11",
        b"id": b"00:00:01:00:00:02",
        b"md": b"unittest",
        b"ci": b"5",
        b"sf": b"0",
        b"ff": b"1",
    }
    info = AsyncServiceInfo(
        "_hap._tcp.local.",
        "foo2._hap._tcp.local.",
        addresses=[socket.inet_aton("127.0.0.1")],
        port=1234,
        properties=desc,
        weight=0,
        priority=0,
    )
    svc = HomeKitService.from_service_info(info)

    assert svc.state_num == 0


def test_missing_sf():
    desc = {
        b"c#": b"11",
        b"id": b"00:00:01:00:00:02",
        b"md": b"unittest",
        b"ci": b"5",
        b"s#": b"15",
        b"ff": b"1",
    }
    info = AsyncServiceInfo(
        "_hap._tcp.local.",
        "foo2._hap._tcp.local.",
        addresses=[socket.inet_aton("127.0.0.1")],
        port=1234,
        properties=desc,
        weight=0,
        priority=0,
    )
    svc = HomeKitService.from_service_info(info)

    assert svc.status_flags == 0


def test_missing_ff():
    desc = {
        b"c#": b"11",
        b"id": b"00:00:01:00:00:02",
        b"md": b"unittest",
        b"ci": b"5",
        b"s#": b"15",
        b"sf": b"0",
    }
    info = AsyncServiceInfo(
        "_hap._tcp.local.",
        "foo2._hap._tcp.local.",
        addresses=[socket.inet_aton("127.0.0.1")],
        port=1234,
        properties=desc,
        weight=0,
        priority=0,
    )
    svc = HomeKitService.from_service_info(info)

    assert svc.feature_flags == 0


def test_missing_md():
    desc = {
        b"c#": b"11",
        b"id": b"00:00:01:00:00:02",
        b"ci": b"5",
        b"s#": b"15",
        b"sf": b"0",
        b"ff": b"44",
    }
    info = AsyncServiceInfo(
        "_hap._tcp.local.",
        "foo2._hap._tcp.local.",
        addresses=[socket.inet_aton("127.0.0.1")],
        port=1234,
        properties=desc,
        weight=0,
        priority=0,
    )
    svc = HomeKitService.from_service_info(info)

    assert svc.model == ""


def test_missing_ci():
    desc = {
        b"c#": b"11",
        b"id": b"00:00:01:00:00:02",
        b"md": b"unittest",
        b"s#": b"15",
        b"sf": b"0",
        b"ff": b"44",
    }
    info = AsyncServiceInfo(
        "_hap._tcp.local.",
        "foo2._hap._tcp.local.",
        addresses=[socket.inet_aton("127.0.0.1")],
        port=1234,
        properties=desc,
        weight=0,
        priority=0,
    )
    svc = HomeKitService.from_service_info(info)

    assert svc.category == Categories.OTHER


def test_missing_id():
    desc = {
        b"c#": b"11",
        b"md": b"unittest",
        b"s#": b"15",
        b"sf": b"0",
        b"ff": b"44",
        b"ci": b"0",
    }
    info = AsyncServiceInfo(
        "_hap._tcp.local.",
        "foo2._hap._tcp.local.",
        addresses=[socket.inet_aton("127.0.0.1")],
        port=1234,
        properties=desc,
        weight=0,
        priority=0,
    )

    with pytest.raises(ValueError):
        HomeKitService.from_service_info(info)


def test_ignore_link_local():
    desc = {
        b"c#": b"1",
        b"id": b"00:00:01:00:00:02",
        b"md": b"unittest",
        b"s#": b"11",
        b"ci": b"5",
        b"sf": b"0",
        b"ff": b"1",
    }
    info = AsyncServiceInfo(
        "_hap._tcp.local.",
        "foo2._hap._tcp.local.",
        addresses=[socket.inet_aton("169.254.121.37"), socket.inet_aton("127.0.0.1")],
        port=1234,
        properties=desc,
        weight=0,
        priority=0,
    )
    svc = HomeKitService.from_service_info(info)

    assert svc.name == "foo2"
    assert svc.type == "_hap._tcp.local."
    assert svc.id == "00:00:01:00:00:02"
    assert svc.model == "unittest"
    assert svc.config_num == 1
    assert svc.state_num == 11
    assert svc.status_flags == 0
    assert svc.feature_flags == FeatureFlags.SUPPORTS_APPLE_AUTHENTICATION_COPROCESSOR
    assert svc.category == Categories.LIGHTBULB
    assert svc.address == "127.0.0.1"
    assert svc.addresses == ["127.0.0.1"]
    assert svc.port == 1234


def test_ignore_link_local_ipv6():
    desc = {
        b"c#": b"1",
        b"id": b"00:00:01:00:00:02",
        b"md": b"unittest",
        b"s#": b"11",
        b"ci": b"5",
        b"sf": b"0",
        b"ff": b"1",
    }
    info = AsyncServiceInfo(
        "_hap._tcp.local.",
        "foo2._hap._tcp.local.",
        addresses=[
            socket.inet_aton("169.254.121.37"),
            socket.inet_pton(socket.AF_INET6, "2a00:1450:4009:820::200e"),
        ],
        port=1234,
        properties=desc,
        weight=0,
        priority=0,
    )
    svc = HomeKitService.from_service_info(info)

    assert svc.name == "foo2"
    assert svc.type == "_hap._tcp.local."
    assert svc.id == "00:00:01:00:00:02"
    assert svc.model == "unittest"
    assert svc.config_num == 1
    assert svc.state_num == 11
    assert svc.status_flags == 0
    assert svc.feature_flags == FeatureFlags.SUPPORTS_APPLE_AUTHENTICATION_COPROCESSOR
    assert svc.category == Categories.LIGHTBULB
    assert svc.address == "2a00:1450:4009:820::200e"
    assert svc.addresses == ["2a00:1450:4009:820::200e"]
    assert svc.port == 1234


def test_prefer_ipv4():
    desc = {
        b"c#": b"1",
        b"id": b"00:00:01:00:00:02",
        b"md": b"unittest",
        b"s#": b"11",
        b"ci": b"5",
        b"sf": b"0",
        b"ff": b"1",
    }
    info = AsyncServiceInfo(
        "_hap._tcp.local.",
        "foo2._hap._tcp.local.",
        addresses=[
            socket.inet_pton(socket.AF_INET6, "2a00:1450:4009:820::200e"),
            socket.inet_aton("127.0.0.1"),
        ],
        port=1234,
        properties=desc,
        weight=0,
        priority=0,
    )
    svc = HomeKitService.from_service_info(info)

    assert svc.name == "foo2"
    assert svc.type == "_hap._tcp.local."
    assert svc.id == "00:00:01:00:00:02"
    assert svc.model == "unittest"
    assert svc.config_num == 1
    assert svc.state_num == 11
    assert svc.status_flags == 0
    assert svc.feature_flags == FeatureFlags.SUPPORTS_APPLE_AUTHENTICATION_COPROCESSOR
    assert svc.category == Categories.LIGHTBULB
    assert svc.address == "127.0.0.1"
    assert svc.addresses == ["127.0.0.1", "2a00:1450:4009:820::200e"]
    assert svc.port == 1234


def test_ignore_unspecified():
    desc = {
        b"c#": b"1",
        b"id": b"00:00:01:00:00:02",
        b"md": b"unittest",
        b"s#": b"11",
        b"ci": b"5",
        b"sf": b"0",
        b"ff": b"1",
    }
    info = AsyncServiceInfo(
        "_hap._tcp.local.",
        "foo2._hap._tcp.local.",
        addresses=[socket.inet_aton("0.0.0.0"), socket.inet_aton("127.0.0.1")],
        port=1234,
        properties=desc,
        weight=0,
        priority=0,
    )
    svc = HomeKitService.from_service_info(info)

    assert svc.name == "foo2"
    assert svc.type == "_hap._tcp.local."
    assert svc.id == "00:00:01:00:00:02"
    assert svc.model == "unittest"
    assert svc.config_num == 1
    assert svc.state_num == 11
    assert svc.status_flags == 0
    assert svc.feature_flags == FeatureFlags.SUPPORTS_APPLE_AUTHENTICATION_COPROCESSOR
    assert svc.category == Categories.LIGHTBULB
    assert svc.address == "127.0.0.1"
    assert svc.addresses == ["127.0.0.1"]
    assert svc.port == 1234


def test_ignore_unspecified_ipv6():
    desc = {
        b"c#": b"1",
        b"id": b"00:00:01:00:00:02",
        b"md": b"unittest",
        b"s#": b"11",
        b"ci": b"5",
        b"sf": b"0",
        b"ff": b"1",
    }
    info = AsyncServiceInfo(
        "_hap._tcp.local.",
        "foo2._hap._tcp.local.",
        addresses=[
            socket.inet_aton("0.0.0.0"),
            socket.inet_pton(socket.AF_INET6, "2a00:1450:4009:820::200e"),
        ],
        port=1234,
        properties=desc,
        weight=0,
        priority=0,
    )
    svc = HomeKitService.from_service_info(info)

    assert svc.name == "foo2"
    assert svc.type == "_hap._tcp.local."
    assert svc.id == "00:00:01:00:00:02"
    assert svc.model == "unittest"
    assert svc.config_num == 1
    assert svc.state_num == 11
    assert svc.status_flags == 0
    assert svc.feature_flags == FeatureFlags.SUPPORTS_APPLE_AUTHENTICATION_COPROCESSOR
    assert svc.category == Categories.LIGHTBULB
    assert svc.address == "2a00:1450:4009:820::200e"
    assert svc.addresses == ["2a00:1450:4009:820::200e"]
    assert svc.port == 1234


def test_no_valid_addresses():
    desc = {
        b"c#": b"1",
        b"id": b"00:00:01:00:00:02",
        b"md": b"unittest",
        b"s#": b"11",
        b"ci": b"5",
        b"sf": b"0",
        b"ff": b"1",
    }
    info = AsyncServiceInfo(
        "_hap._tcp.local.",
        "foo2._hap._tcp.local.",
        addresses=[
            socket.inet_aton("0.0.0.0"),
        ],
        port=1234,
        properties=desc,
        weight=0,
        priority=0,
    )
    with pytest.raises(
        ValueError,
        match="Invalid HomeKit Zeroconf record: Missing non-link-local or unspecified address",
    ):
        HomeKitService.from_service_info(info)



#
# Copyright 2019 aiohomekit team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import pytest

from aiohomekit.http import HttpStatusCodes


def test_1():
    assert HttpStatusCodes[HttpStatusCodes.INTERNAL_SERVER_ERROR] == "Internal Server Error"


def test_unknown_code():
    with pytest.raises(KeyError):
        HttpStatusCodes[99]



from __future__ import annotations

import pytest

from aiohomekit.controller.ble.manufacturer_data import HomeKitAdvertisement


def test_manufacturer_data_too_short() -> None:
    """Test that short manufacturer data raises appropriate error."""
    # Valid HomeKit advertisement type but buffer too short (less than 15 bytes)
    short_data = b"\x06\x00\x00\x01\x02\x03\x04\x05\x06"  # Only 9 bytes

    with pytest.raises(ValueError, match="HomeKit advertisement data too short"):
        HomeKitAdvertisement.from_manufacturer_data(
            name="Test", address="00:00:00:00:00:00", manufacturer_data={76: short_data}
        )


def test_manufacturer_data_without_setup_hash() -> None:
    """Test with 15 bytes (no setup hash - valid for older devices)."""
    data_no_hash = b"\x06\x00\x00\x01\x02\x03\x04\x05\x06\x07\x08\x00\x00\x00\x00"  # 15 bytes total

    result = HomeKitAdvertisement.from_manufacturer_data(
        name="Test", address="00:00:00:00:00:00", manufacturer_data={76: data_no_hash}
    )

    assert result is not None
    assert result.address == "00:00:00:00:00:00"
    assert result.name == "Test"
    assert result.setup_hash == b""  # No setup hash


def test_manufacturer_data_with_setup_hash() -> None:
    """Test with exactly 19 bytes (includes setup hash)."""
    valid_data = (
        b"\x06\x00\x00\x01\x02\x03\x04\x05\x06\x07\x08\x00\x00\x00\x00\x00\x11\x22\x33"  # 19 bytes total
    )

    result = HomeKitAdvertisement.from_manufacturer_data(
        name="Test", address="00:00:00:00:00:00", manufacturer_data={76: valid_data}
    )

    assert result is not None
    assert result.address == "00:00:00:00:00:00"
    assert result.name == "Test"
    assert result.setup_hash == b"\x00\x11\x22\x33"  # Has setup hash



#
# Copyright 2022 aiohomekit team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import pytest

from aiohomekit.pdu import (
    OpCode,
    PDUStatus,
    decode_pdu,
    decode_pdu_continuation,
    encode_pdu,
)


def test_encode():
    assert list(encode_pdu(OpCode.CHAR_SIG_READ, 55, 1)) == [b"\x00\x017\x01\x00"]


def test_encode_with_body():
    assert list(encode_pdu(OpCode.CHAR_SIG_READ, 44, 1, b"SOMEDATA")) == [
        b"\x00\x01,\x01\x00\x08\x00SOMEDATA"
    ]


def test_encode_with_fragments():
    result = list(encode_pdu(OpCode.CHAR_SIG_READ, 44, 1, b"ABCD" * 64, fragment_size=256))

    assert result == [
        b"\x00\x01,\x01\x00\x00\x01ABCDABCDABCDABCDABCDABCDABCDABCDABCDABCDABCDABCDA"
        b"BCDABCDABCDABCDABCDABCDABCDABCDABCDABCDABCDABCDABCDABCDABCDABCDABCDABCDABCDA"
        b"BCDABCDABCDABCDABCDABCDABCDABCDABCDABCDABCDABCDABCDABCDABCDABCDABCDABCDABCDA"
        b"BCDABCDABCDABCDABCDABCDABCDABCDABCDABCDABCDABCDA",
        b"\x80,BCDABCD",
    ]


def test_decode():
    assert decode_pdu(23, b"\x00\x17\x00") == (PDUStatus.SUCCESS, 0, b"")


def test_decode_with_body():
    assert decode_pdu(23, b"\x00\x17\x00\x08\x00SOMEDATA") == (
        PDUStatus.SUCCESS,
        8,
        b"SOMEDATA",
    )


def test_decode_invalid_tid():
    with pytest.raises(ValueError):
        decode_pdu(24, b"\x00\x17\x00")


def test_decode_invalid_status():
    assert decode_pdu(23, b"\x00\x17\x01") == (PDUStatus.UNSUPPORTED_PDU, 0, b"")


def test_decode_continuation():
    assert decode_pdu_continuation(23, b"\x80\x17SOMEDATA") == b"SOMEDATA"


def test_decode_continuation_invalid_cbit():
    with pytest.raises(ValueError):
        decode_pdu_continuation(23, b"\x00\x17SOMEDATA")


def test_decode_continuation_invalid_tid():
    with pytest.raises(ValueError):
        decode_pdu_continuation(24, b"\x80\x17SOMEDATA")



"""Test status codes."""

from aiohomekit.protocol.statuscodes import HapStatusCode, to_status_code


async def test_normalized_statuscodes():
    """Verify we account for quirks in HAP implementations."""
    assert to_status_code(70411) == HapStatusCode.INSUFFICIENT_AUTH
    assert to_status_code(-70411) == HapStatusCode.INSUFFICIENT_AUTH



from unittest.mock import AsyncMock, MagicMock, patch

import pytest

from aiohomekit.controller import Controller, controller as controller_module
from aiohomekit.controller.abstract import TransportType
from aiohomekit.controller.ble.controller import BleController
from aiohomekit.controller.ip.controller import IpController
from aiohomekit.exceptions import AccessoryDisconnectedError


async def test_remove_pairing(controller_and_paired_accessory):
    pairing = controller_and_paired_accessory.aliases["alias"]

    # Verify that there is a pairing connected and working
    await pairing.get_characteristics([(1, 9)])

    # Remove pairing from controller
    await controller_and_paired_accessory.remove_pairing("alias")

    # Verify now gives an appropriate error
    with pytest.raises(AccessoryDisconnectedError):
        await pairing.get_characteristics([(1, 9)])


async def test_passing_in_bleak_to_controller():
    """Test we can pass in a bleak scanner instance to the controller.

    Passing in the instance should enable BLE scanning.
    """
    with (
        patch.object(controller_module, "BLE_TRANSPORT_SUPPORTED", False),
        patch.object(controller_module, "COAP_TRANSPORT_SUPPORTED", False),
        patch.object(controller_module, "IP_TRANSPORT_SUPPORTED", False),
    ):
        controller = Controller(bleak_scanner_instance=AsyncMock(register_detection_callback=MagicMock()))
        await controller.async_start()

    assert len(controller.transports) == 1
    assert isinstance(controller.transports[TransportType.BLE], BleController)


async def test_passing_in_async_zeroconf(mock_asynczeroconf):
    """Test we can pass in a zeroconf ServiceBrowser instance to the controller.

    Passing in the instance should enable zeroconf scanning.
    """
    with (
        patch.object(controller_module, "BLE_TRANSPORT_SUPPORTED", False),
        patch.object(controller_module, "COAP_TRANSPORT_SUPPORTED", False),
        patch.object(controller_module, "IP_TRANSPORT_SUPPORTED", False),
    ):
        controller = Controller(async_zeroconf_instance=mock_asynczeroconf)
        await controller.async_start()

    assert len(controller.transports) == 1
    assert isinstance(controller.transports[TransportType.IP], IpController)
